@@ -8,13 +8,13 @@ implementation.
 
 # Copyright (C) 2020 The Psycopg Team
 
+import sys
 import logging
 from os import getpid
 from weakref import ref
-from functools import partial
 
-from ctypes import Array, pointer, string_at, create_string_buffer, byref
-from ctypes import addressof, c_char_p, c_int, c_size_t, c_ulong
+from ctypes import Array, POINTER, cast, string_at, create_string_buffer, byref
+from ctypes import addressof, c_char_p, c_int, c_size_t, c_ulong, c_void_p, py_object
 from typing import Any, Callable, List, Optional, Sequence, Tuple
 from typing import cast as t_cast, TYPE_CHECKING
 
@@ -22,7 +22,7 @@ from .. import errors as e
 from . import _pq_ctypes as impl
 from .misc import PGnotify, ConninfoOption, PGresAttDesc
 from .misc import error_message, connection_summary
-from ._enums import Format, ExecStatus
+from ._enums import Format, ExecStatus, Trace
 
 # Imported locally to call them from __del__ methods
 from ._pq_ctypes import PQclear, PQfinish, PQfreeCancel, PQstatus
@@ -45,10 +45,9 @@ def version() -> int:
     return impl.PQlibVersion()
 
 
-def notice_receiver(
-    arg: Any, result_ptr: impl.PGresult_struct, wconn: "ref[PGconn]"
-) -> None:
-    pgconn = wconn()
+@impl.PQnoticeReceiver  # type: ignore
+def notice_receiver(arg: c_void_p, result_ptr: impl.PGresult_struct) -> None:
+    pgconn = cast(arg, POINTER(py_object)).contents.value()
     if not (pgconn and pgconn.notice_handler):
         return
 
@@ -57,8 +56,8 @@ def notice_receiver(
         pgconn.notice_handler(res)
     except Exception as exc:
         logger.exception("error in notice receiver: %s", exc)
-
-    res._pgresult_ptr = None  # avoid destroying the pgresult_ptr
+    finally:
+        res._pgresult_ptr = None  # avoid destroying the pgresult_ptr
 
 
 class PGconn:
@@ -70,7 +69,7 @@ class PGconn:
         "_pgconn_ptr",
         "notice_handler",
         "notify_handler",
-        "_notice_receiver",
+        "_self_ptr",
         "_procpid",
         "__weakref__",
     )
@@ -80,10 +79,9 @@ class PGconn:
         self.notice_handler: Optional[Callable[["abc.PGresult"], None]] = None
         self.notify_handler: Optional[Callable[[PGnotify], None]] = None
 
-        self._notice_receiver = impl.PQnoticeReceiver(  # type: ignore
-            partial(notice_receiver, wconn=ref(self))
-        )
-        impl.PQsetNoticeReceiver(pgconn_ptr, self._notice_receiver, None)
+        # Keep alive for the lifetime of PGconn
+        self._self_ptr = py_object(ref(self))
+        impl.PQsetNoticeReceiver(pgconn_ptr, notice_receiver, byref(self._self_ptr))
 
         self._procpid = getpid()
 
@@ -239,10 +237,19 @@ class PGconn:
 
     @property
     def needs_password(self) -> bool:
+        """True if the connection authentication method required a password,
+        but none was available.
+
+        See :pq:`PQconnectionNeedsPassword` for details.
+        """
         return bool(impl.PQconnectionNeedsPassword(self._pgconn_ptr))
 
     @property
     def used_password(self) -> bool:
+        """True if the connection authentication method used a password.
+
+        See :pq:`PQconnectionUsedPassword` for details.
+        """
         return bool(impl.PQconnectionUsedPassword(self._pgconn_ptr))
 
     @property
@@ -529,6 +536,7 @@ class PGconn:
             )
 
     def flush(self) -> int:
+        # PQflush segfaults if it receives a NULL connection
         if not self._pgconn_ptr:
             raise e.OperationalError("flushing failed: the connection is closed")
         rv: int = impl.PQflush(self._pgconn_ptr)
@@ -589,9 +597,43 @@ class PGconn:
         else:
             return nbytes, memoryview(b"")
 
+    def trace(self, fileno: int) -> None:
+        """
+        Enable tracing of the client/server communication to a file stream.
+
+        See :pq:`PQtrace` for details.
+        """
+        if sys.platform != "linux":
+            raise e.NotSupportedError("currently only supported on Linux")
+        stream = impl.fdopen(fileno, b"w")
+        impl.PQtrace(self._pgconn_ptr, stream)
+
+    def set_trace_flags(self, flags: Trace) -> None:
+        """
+        Configure tracing behavior of client/server communication.
+
+        :param flags: operating mode of tracing.
+
+        See :pq:`PQsetTraceFlags` for details.
+        """
+        impl.PQsetTraceFlags(self._pgconn_ptr, flags)
+
+    def untrace(self) -> None:
+        """
+        Disable tracing, previously enabled through `trace()`.
+
+        See :pq:`PQuntrace` for details.
+        """
+        impl.PQuntrace(self._pgconn_ptr)
+
     def encrypt_password(
         self, passwd: bytes, user: bytes, algorithm: Optional[bytes] = None
     ) -> bytes:
+        """
+        Return the encrypted form of a PostgreSQL password.
+
+        See :pq:`PQencryptPasswordConn` for details.
+        """
         out = impl.PQencryptPasswordConn(self._pgconn_ptr, passwd, user, algorithm)
         if not out:
             raise e.OperationalError(
@@ -841,7 +883,11 @@ class PGcancel:
         See :pq:`PQcancel()` for details.
         """
         buf = create_string_buffer(256)
-        res = impl.PQcancel(self.pgcancel_ptr, pointer(buf), len(buf))  # type: ignore
+        res = impl.PQcancel(
+            self.pgcancel_ptr,
+            byref(buf),  # type: ignore[arg-type]
+            len(buf),
+        )
         if not res:
             raise e.OperationalError(
                 f"cancel failed: {buf.value.decode('utf8', 'ignore')}"
@@ -869,7 +915,7 @@ class Conninfo:
             raise TypeError(f"bytes expected, got {type(conninfo)} instead")
 
         errmsg = c_char_p()
-        rv = impl.PQconninfoParse(conninfo, pointer(errmsg))
+        rv = impl.PQconninfoParse(conninfo, byref(errmsg))  # type: ignore[arg-type]
         if not rv:
             if not errmsg:
                 raise MemoryError("couldn't allocate on conninfo parse")
@@ -953,10 +999,10 @@ class Escaping:
             out = create_string_buffer(len(data) * 2 + 1)
             impl.PQescapeStringConn(
                 self.conn._pgconn_ptr,
-                pointer(out),  # type: ignore
+                byref(out),  # type: ignore[arg-type]
                 data,
                 len(data),
-                pointer(error),
+                byref(error),  # type: ignore[arg-type]
             )
 
             if error:
@@ -967,7 +1013,7 @@ class Escaping:
         else:
             out = create_string_buffer(len(data) * 2 + 1)
             impl.PQescapeString(
-                pointer(out),  # type: ignore
+                byref(out),  # type: ignore[arg-type]
                 data,
                 len(data),
             )
@@ -987,10 +1033,14 @@ class Escaping:
                 self.conn._pgconn_ptr,
                 data,
                 len(data),
-                pointer(t_cast(c_ulong, len_out)),
+                byref(t_cast(c_ulong, len_out)),  # type: ignore[arg-type]
             )
         else:
-            out = impl.PQescapeBytea(data, len(data), pointer(t_cast(c_ulong, len_out)))
+            out = impl.PQescapeBytea(
+                data,
+                len(data),
+                byref(t_cast(c_ulong, len_out)),  # type: ignore[arg-type]
+            )
         if not out:
             raise MemoryError(
                 f"couldn't allocate for escape_bytea of {len(data)} bytes"
@@ -1007,7 +1057,10 @@ class Escaping:
             self.conn._ensure_pgconn()
 
         len_out = c_size_t()
-        out = impl.PQunescapeBytea(data, pointer(t_cast(c_ulong, len_out)))
+        out = impl.PQunescapeBytea(
+            data,
+            byref(t_cast(c_ulong, len_out)),  # type: ignore[arg-type]
+        )
         if not out:
             raise MemoryError(
                 f"couldn't allocate for unescape_bytea of {len(data)} bytes"
@@ -1024,3 +1077,5 @@ import ssl  # noqa
 # disable libcrypto setup in libpq, so it won't stomp on the callbacks
 # that have already been set up
 impl.PQinitOpenSSL(1, 0)
+
+__build_version__ = version()
